import json
from abc import ABC
from typing import Dict, Union

from pydantic import ValidationError

from tktl.core.exceptions import NoContentsFoundException, UserRepoValidationException
from tktl.core.schemas.project import ProjectValidationOutput


class ValidationOutputBase(ABC):
    title: str
    summary: str
    text: str

    @classmethod
    def format_outputs(cls, outputs: Dict):
        if outputs:
            text = cls.text + "\n## Outputs\n"
            for k, v in outputs.items():
                text += f"- {k}: {v}\n"
            return text
        else:
            return cls.text

    @classmethod
    def get_check_created_output(cls):
        return ProjectValidationOutput(title=cls.title, summary=cls.summary, text="")

    @classmethod
    def default_output(cls, passed: bool, debug_output: str):
        return f"""
Debug Output
{debug_output}
        """


class ConfigFileValidationFailure(ValidationOutputBase):
    title: str = """Invalid Config File ❌"""
    summary: str = (
        "We've identified erros in your configuration file (tktl.yaml)."
        " See below for details. Pushes will not succeed or result in new builds"
    )
    text = """"""

    @classmethod
    def format_step_results(cls, validation_errors: ValidationError):
        debug_output = ""
        for error in json.loads(validation_errors.json()):
            debug_output += f"""
- Value for: {': '.join(f'`{i}`' for i in error['loc'])} is invalid: `{error['msg']}`
"""
        default_output = cls.default_output(passed=False, debug_output=debug_output)
        return default_output


class ProjectValidationFailure(ValidationOutputBase):
    title: str = """Missing Files or Directory ❌"""
    summary: str = (
        "We've identified invalid configuration of your repository."
        " See below for details. Pushes will not succeed or result in new builds"
    )
    text = """"""

    @classmethod
    def format_step_results(
        cls,
        validation_errors: Union[UserRepoValidationException, NoContentsFoundException],
    ):
        debug_output = f"""
Missing files in repository: {' '.join(validation_errors.missing_files) if
        validation_errors.missing_files else '`NA`'}
Missing directories in repository: {' '.join(validation_errors.missing_directories) if
        validation_errors.missing_directories else '`NA`'}
tktl.yaml config file present: {'❌' if validation_errors.missing_config else '✅'}
        """
        default_output = cls.default_output(passed=False, debug_output=debug_output)

        return default_output



import click
from taktile_types.enums.instance import ServiceType

from tktl.cli.common import ClickCommand
from tktl.commands.health import check_grpc_health, check_rest_health
from tktl.core.config import settings
from tktl.core.exceptions import APIClientException, TaktileSdkError
from tktl.core.loggers import LOG


@click.command(
    "health",
    help="Health check your endpoints",
    cls=ClickCommand,
    **settings.HELP_COLORS_DICT
)
@click.option("-r", "--repo", help="Repository owner & name: owner/repo-name")
@click.option("-b", "--branch", help="Branch name")
@click.option(
    "-s",
    "--service",
    help="Service kind",
    type=click.Choice([ServiceType.REST, ServiceType.GRPC]),
    default=ServiceType.REST,
)
@click.option(
    "-l", "--local", help="Run against local endpoint", is_flag=True, default=False
)
@click.pass_context
def health(
    ctx,
    repo: str,
    branch: str,
    service: str = ServiceType.REST,
    local: bool = False,
):
    if not local:
        if not repo and not branch:
            return click.echo(health.get_help(ctx=ctx))
        if (repo and not branch) or (branch and not repo):
            LOG.error("If not running locally, must set branch and repo name")
            return
    try:
        check_rest_health() if service == ServiceType.REST else check_grpc_health()
    except APIClientException:
        exit(1)
    except TaktileSdkError as e:
        LOG.error(str(e))



from typing import List

from tktl.commands import BaseDeploymentApiCommand
from tktl.core.config import settings
from tktl.core.schemas.repository import Repository


class GetRepositories(BaseDeploymentApiCommand):
    def execute(self) -> List[Repository]:
        return self.client.call(
            verb="get", path=f"{settings.API_V1_STR}/models", model=Repository
        )



"""Template demonstrating Feature Monitoring.
Documentation https://docs.taktile.com/using-taktile/monitoring
"""
from pydantic import BaseModel

from tktl.future import Tktl
from tktl.future.monitor import Tracker

# instantiate client
client = Tktl()


class Payload(BaseModel):
    feature: str


class Response(BaseModel):
    result: str


# endpoints
@client.endpoint(
    X=Payload, y=Response, track_inputs=["feature"], track_outputs=["result"]
)
def func(x: Payload, tracker: Tracker) -> Response:
    tracker.log_numerical("feature_len", len(x.feature))
    return Response(result=x.feature)



from pyarrow.flight import (  # type: ignore
    FlightCancelledError,
    FlightClient,
    FlightUnauthenticatedError,
    FlightUnavailableError,
)

from tktl.core.clients import API
from tktl.core.config import settings as tktl_settings
from tktl.core.exceptions import APIClientException


def check_rest_health():
    client = API(api_base=tktl_settings.LOCAL_REST_ENDPOINT)
    assert client.call(verb="get", path="healthz", raw=True).status_code == 204


def check_grpc_health():
    try:
        client = FlightClient(
            tls_root_certs=None,
            location=tktl_settings.LOCAL_ARROW_ENDPOINT,
        )
        client.wait_for_available(timeout=1)
    except FlightUnauthenticatedError:
        return True
    except (FlightCancelledError, FlightUnavailableError) as e:
        error_str = f"Service is not running properly: {repr(e.detail)}"
        raise APIClientException(detail=error_str, status_code=e.status_code)



import click
from pydantic import UUID4

from tktl.cli.common import ClickGroup, to_uuid
from tktl.commands.deployments import DeleteDeployment
from tktl.core.config import settings
from tktl.core.exceptions import APIClientException
from tktl.core.loggers import LOG
from tktl.login import validate_decorator


@click.group(
    "delete", help="Delete resources", cls=ClickGroup, **settings.HELP_COLORS_DICT
)
def delete():
    pass


@delete.command("deployment", help="Manage deployments")
@click.argument("deployment_id", required=True, callback=to_uuid)
@validate_decorator
def delete_deployment_by_id(deployment_id: UUID4):
    inp = click.prompt(
        f"Deleting deployment with id {deployment_id}. Do you wish to continue? [y (or press enter)/n]",
        default="y",
        show_choices=False,
        show_default=False,
    )
    if inp != "y":
        LOG.error("Safe choice!")
    command = DeleteDeployment()
    try:
        response = command.execute(deployment_id=deployment_id)
    except APIClientException:
        LOG.error("Only super users have access to this method at this time")
        return
    LOG.warning(f"Deployment with id: {response.id}")
    return response



from datetime import datetime
from typing import Dict, List, Optional, Union

import certifi
from pydantic import UUID4, BaseModel, validator
from taktile_types.enums.instance import ServiceType
from taktile_types.enums.repository.access import AccessKind

from tktl.constants import NON_EXISTING_DEPLOYMENT_DEFAULT_STATUS, URL_UNAVAILABLE
from tktl.core.config import settings


class TablePrintableBaseModelMixin:
    def table_repr(self, subset: List[str] = None) -> Dict:
        ...


class Endpoint(BaseModel):
    name: str
    kind: str
    deployment_id: Optional[UUID4]
    profiling_supported: bool
    has_rest_sample_data: Optional[bool] = True
    has_arrow_sample_data: Optional[bool] = True

    class Config:
        validate_assignment = True

    @validator("deployment_id")
    def set_deployment_id(cls, deployment_id: UUID4):
        return deployment_id

    def table_repr(self, subset=None):
        as_dict = self.dict()
        as_dict["NAME"] = as_dict.pop("name")
        as_dict["KIND"] = str(as_dict.pop("kind"))
        as_dict["PROFILING SUPPORTED"] = str(as_dict.pop("profiling_supported"))
        as_dict["DEPLOYMENT ID"] = str(as_dict.pop("deployment_id"))
        return as_dict


class UserRepository(BaseModel):
    id: UUID4
    full_name: str


class RepositorySecret(BaseModel):
    id: UUID4
    repository_id: UUID4
    secret_name: str


class RepositoryDeployment(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    created_at: datetime
    status: Optional[str]
    public_docs_url: Optional[str]

    rest_cpu_request: Optional[str]
    rest_memory_request: Optional[str]
    rest_gpu_request: Optional[str]
    rest_replicas: Optional[int]
    max_rest_replicas: Optional[int]

    arrow_replicas: Optional[int]
    arrow_cpu_request: Optional[str]
    arrow_memory_request: Optional[str]
    arrow_gpu_request: Optional[str]

    git_ref: str
    commit_hash: str
    n_endpoints: Optional[int]

    @validator("status", always=True)
    def validate_status(cls, value):
        return value if value else NON_EXISTING_DEPLOYMENT_DEFAULT_STATUS

    @validator("n_endpoints", always=True)
    def validate_n_endpoints(cls, value):
        return value or 0

    def table_repr(self, subset=None):
        as_dict = self.dict(
            exclude={
                "service_type",
                "endpoints",
            }
        )
        repr = {}
        repr["ID"] = str(as_dict["id"])
        repr["BRANCH"] = as_dict["git_ref"]
        repr["COMMIT"] = as_dict["commit_hash"][0:7]
        repr["STATUS"] = as_dict["status"]
        repr["ENDPOINTS"] = as_dict["n_endpoints"]
        repr["CREATED AT"] = str(as_dict["created_at"])
        repr["REST DOCS URL"] = _format_http_url(as_dict["public_docs_url"])

        repr[
            "REST REPLICAS"
        ] = f"{as_dict['max_rest_replicas']} ({as_dict['rest_replicas']})"
        repr["REST SIZE"] = self.get_requests(ServiceType.REST, as_dict)

        if as_dict["arrow_replicas"] != "0":
            repr["ARROW REPLICAS"] = as_dict["arrow_replicas"]
            repr["ARROW SIZE"] = self.get_requests(ServiceType.GRPC, as_dict)
        if subset:
            return {k: v for k, v in repr.items() if k in subset}
        return repr

    def get_requests(self, kind: ServiceType, values: Dict):
        kind_name = "rest" if kind == ServiceType.REST else "arrow"

        if values[f"{kind_name}_cpu_request"] is None:
            return ""
        requests = f"{values[f'{kind_name}_cpu_request']} {values[f'{kind_name}_memory_request']}"
        if values[f"{kind_name}_gpu_request"] != "0":
            return requests + f" {values[f'{kind_name}_gpu_request']} gpus"
        return requests


class Repository(BaseModel, TablePrintableBaseModelMixin):
    id: UUID4
    ref_id: int
    repository_name: str
    repository_owner: str
    repository_description: Optional[str] = None
    access: AccessKind
    deployments: List[RepositoryDeployment]
    n_deployments: Optional[int] = None

    @validator("n_deployments", always=True)
    def validate_n_deployments(cls, _, values):
        return len(values["deployments"])

    def __hash__(self):
        return self.id.__hash__()

    def table_repr(self, subset=None):
        as_dict = self.dict(exclude={"ref_id", "deployments"})
        as_dict["ID"] = f"{as_dict.pop('id')}"
        as_dict[
            "FULL NAME"
        ] = f"{as_dict.pop('repository_owner')}/{as_dict.pop('repository_name')}"
        as_dict["DEPLOYMENTS"] = as_dict.pop("n_deployments")
        as_dict["ACCESS"] = f"{as_dict.pop('access').value}"
        desc = as_dict.pop("repository_description")
        as_dict["DESCRIPTION"] = f"{desc[0:20] + '...' if desc else '-'}"
        if subset:
            return {k: v for k, v in as_dict.items() if k in subset}
        return as_dict


class ReportResponse(BaseModel):
    deployment_id: UUID4
    endpoint_name: str
    report_type: str
    chart_name: Optional[str] = None
    variable_name: Optional[str] = None
    value: Union[List, Dict]


def _format_http_url(url: Optional[str], docs: bool = True) -> str:
    if url == URL_UNAVAILABLE or url is None:
        return URL_UNAVAILABLE
    if settings.LOCAL_STACK:
        return f"http://{url}:8000/{'docs' if docs else ''}"
    if settings.E2E:
        return f"http://{url}/{'docs' if docs else ''}".replace(
            ".local.taktile.com", "-rest.default.svc.cluster.local", 1
        )

    return f"https://{url}/{'docs' if docs else ''}"


def _format_grpc_url(url: Optional[str]) -> str:
    if url == URL_UNAVAILABLE or url is None:
        return URL_UNAVAILABLE
    if settings.LOCAL_STACK:
        return (
            f"grpc+tcp://{url}:5005"
            if (url and url != "UNAVAILABLE")
            else "UNAVAILABLE"
        )
    if settings.E2E:
        return (
            f"grpc+tcp://{url}:5005".replace(
                ".local.taktile.com", "-grpc.default.svc.cluster.local", 1
            )
            if (url and url != "UNAVAILABLE")
            else "UNAVAILABLE"
        )
    return f"grpc+tls://{url}:5005" if (url and url != "UNAVAILABLE") else "UNAVAILABLE"


def load_certs():
    with open(certifi.where(), "r") as cert:
        return cert.read()


Resources = Union[Repository, RepositoryDeployment, Endpoint, UserRepository]



import typing as t

import click

from tktl.commands import login as login_commands
from tktl.core.exceptions import APIClientException
from tktl.core.loggers import LOG
from tktl.login import login as tktl_login


@click.command("login", help="Log in & store api key")
@click.argument("api_key", required=False)
def login(api_key: t.Optional[str]):

    if not api_key:
        api_key = click.prompt("Please enter your API key", hide_input=True)

    if tktl_login(api_key=api_key):
        command = login_commands.LogInCommand()
        try:
            command.execute()
        except APIClientException as e:
            LOG.error(f"Authentication failed: {e}")


@click.command("logout", help="Log out & remove apiKey from config file")
def logout():
    command = login_commands.LogOutCommand()
    command.execute()



from tktl.core.clients import DeploymentApiClient
from tktl.core.clients.taktile import TaktileApiClient
from tktl.core.loggers import LOG, Logger


class CommandBase(object):
    def __init__(self, api=None):
        self.api = api


class BaseDeploymentApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = DeploymentApiClient(logger=logger)


class BaseTaktileApiCommand:
    def __init__(self, logger: Logger = LOG):
        self.client = TaktileApiClient(logger=logger)



from tktl.core.t import TemplateT

TEMPLATE_PROJECT_DIRS = [
    "src/",
    "assets/",
    "tests/",
]

TEMPLATE_PROJECT_FILES = {
    TemplateT.REPAYMENT: [
        "src/endpoints.py",
        "assets/model.joblib",
        "assets/loans_test.pqt",
        "tests/test_accuracy.py",
        "tests/test_plausibility.py",
        "tests/test_schema.py",
        ".gitignore",
        ".gitattributes",
        ".buildfile",
        ".dockerignore",
        "README.md",
        "requirements.txt",
    ],
    TemplateT.BARE: [
        "src/endpoints.py",
        ".gitignore",
        ".gitattributes",
        ".buildfile",
        ".dockerignore",
        "README.md",
        "requirements.txt",
        "assets/.gitkeep",
    ],
    TemplateT.PRECOMPUTED: [
        "src/endpoints.py",
        "assets/model.joblib",
        "assets/loans_test_X.pqt",
        "assets/loans_test_y.pqt",
        "tests/test_accuracy.py",
        "tests/test_plausibility.py",
        "tests/test_schema.py",
        ".gitignore",
        ".gitattributes",
        ".buildfile",
        ".dockerignore",
        "README.md",
        "requirements.txt",
    ],
    TemplateT.MONITORING: [
        "src/endpoints.py",
        ".gitignore",
        ".gitattributes",
        ".buildfile",
        ".dockerignore",
        "README.md",
        "requirements.txt",
        "assets/.gitkeep",
    ],
}


CONFIG_FILE_TEMPLATE = """
# See the full documentation at https://docs.taktile.com/configuring-taktile/deployment-configuration

# If this option is set only commits with a commit message starting
# with the deployment_prefix will be deployed
# deployment_prefix: "#deploy"

# If this option is set commits with a commit message starting
# with the undeployment_prefix will cause an existing Taktile
# deployment to no longer be deployed
# undeployment_prefix: "#undeploy"

service:

  # Rest deployment scaling parameters
  rest:
    # requests determine the size of each replica. CPU should be specified as 'XXXm', where XXX is
    # thousandths of a core, up to 1000. Memory can be specified in base 2 units (Mi or Gi)
    requests:
      cpu: 500m
      memory: 512Mi
    # Maximum and minimum number of replicas to which the REST service is allowed to reach
    # by the auto-scaler
    max_replicas: 1
    # Minimum, or starting number of replicas
    replicas: 1

  # Arrow deployment scaling parameters
  arrow:
    requests:
      cpu: 500m
      memory: 512Mi
    # Arrow replicas are fixed and don't autoscale
    replicas: 1

version: {version}
"""
